# Flow visualization code used from https://github.com/tomrunia/OpticalFlow_Visualization


# MIT License
#
# Copyright (c) 2018 Tom Runia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-08-03

import numpy as np

def make_colorwheel():
    """
    Generates a color wheel for optical flow visualization as presented in:
        Baker et al. "A Database and Evaluation Methodology for Optical Flow" (ICCV, 2007)
        URL: http://vision.middlebury.edu/flow/flowEval-iccv07.pdf

    Code follows the original C++ source code of Daniel Scharstein.
    Code follows the the Matlab source code of Deqing Sun.

    Returns:
        np.ndarray: Color wheel
    """

    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR
    colorwheel = np.zeros((ncols, 3))
    col = 0

    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.floor(255*np.arange(0,RY)/RY)
    col = col+RY
    # YG
    colorwheel[col:col+YG, 0] = 255 - np.floor(255*np.arange(0,YG)/YG)
    colorwheel[col:col+YG, 1] = 255
    col = col+YG
    # GC
    colorwheel[col:col+GC, 1] = 255
    colorwheel[col:col+GC, 2] = np.floor(255*np.arange(0,GC)/GC)
    col = col+GC
    # CB
    colorwheel[col:col+CB, 1] = 255 - np.floor(255*np.arange(CB)/CB)
    colorwheel[col:col+CB, 2] = 255
    col = col+CB
    # BM
    colorwheel[col:col+BM, 2] = 255
    colorwheel[col:col+BM, 0] = np.floor(255*np.arange(0,BM)/BM)
    col = col+BM
    # MR
    colorwheel[col:col+MR, 2] = 255 - np.floor(255*np.arange(MR)/MR)
    colorwheel[col:col+MR, 0] = 255
    return colorwheel


def flow_uv_to_colors(u, v, convert_to_bgr=False):
    """
    Applies the flow color wheel to (possibly clipped) flow components u and v.

    According to the C++ source code of Daniel Scharstein
    According to the Matlab source code of Deqing Sun

    Args:
        u (np.ndarray): Input horizontal flow of shape [H,W]
        v (np.ndarray): Input vertical flow of shape [H,W]
        convert_to_bgr (bool, optional): Convert output image to BGR. Defaults to False.

    Returns:
        np.ndarray: Flow visualization image of shape [H,W,3]
    """
    flow_image = np.zeros((u.shape[0], u.shape[1], 3), np.uint8)
    colorwheel = make_colorwheel()  # shape [55x3]
    ncols = colorwheel.shape[0]
    rad = np.sqrt(np.square(u) + np.square(v))
    a = np.arctan2(-v, -u)/np.pi
    fk = (a+1) / 2*(ncols-1)
    k0 = np.floor(fk).astype(np.int32)
    k1 = k0 + 1
    k1[k1 == ncols] = 0
    f = fk - k0
    for i in range(colorwheel.shape[1]):
        tmp = colorwheel[:,i]
        col0 = tmp[k0] / 255.0
        col1 = tmp[k1] / 255.0
        col = (1-f)*col0 + f*col1
        idx = (rad <= 1)
        col[idx]  = 1 - rad[idx] * (1-col[idx])
        col[~idx] = col[~idx] * 0.75   # out of range
        # Note the 2-i => BGR instead of RGB
        ch_idx = 2-i if convert_to_bgr else i
        flow_image[:,:,ch_idx] = np.floor(255 * col)
    return flow_image


def flow_to_image(flow_uv, clip_flow=None, convert_to_bgr=False):
    """
    Expects a two dimensional flow image of shape.

    Args:
        flow_uv (np.ndarray): Flow UV image of shape [H,W,2]
        clip_flow (float, optional): Clip maximum of flow values. Defaults to None.
        convert_to_bgr (bool, optional): Convert output image to BGR. Defaults to False.

    Returns:
        np.ndarray: Flow visualization image of shape [H,W,3]
    """
    assert flow_uv.ndim == 3, 'input flow must have three dimensions'
    assert flow_uv.shape[2] == 2, 'input flow must have shape [H,W,2]'
    if clip_flow is not None:
        flow_uv = np.clip(flow_uv, 0, clip_flow)
    u = flow_uv[:,:,0]
    v = flow_uv[:,:,1]
    rad = np.sqrt(np.square(u) + np.square(v))
    rad_max = np.max(rad)
    epsilon = 1e-5
    u = u / (rad_max + epsilon)
    v = v / (rad_max + epsilon)
    return flow_uv_to_colors(u, v, convert_to_bgr)


import tensorflow as tf
import tensorflow_addons as tfa
from tensorflow.keras import layers


def Normalization(norm_type, groups=None):
    if norm_type == 'group':
        return tfa.layers.GroupNormalization(groups)
    elif norm_type == 'batch':
        return layers.BatchNormalization()
    elif norm_type == 'instance':
        return tfa.layers.InstanceNormalization()
    elif norm_type is None:
        return layers.Lambda(lambda x: x)
    else:
        raise ValueError(f'Invalid norm_type specified: {norm_type}')


class ResBlock(layers.Layer):
    def __init__(self, filters, norm_type='group', strides=1, **kwargs):
        super().__init__(**kwargs)
        self.filters = filters
        self.norm_type = norm_type
        self.strides = strides

        self.conv1 = layers.Conv2D(filters, 3, strides, 'same')
        self.conv2 = layers.Conv2D(filters, 3, 1, 'same')

        groups = filters // 8
        self.norm1 = Normalization(norm_type, groups)
        self.norm2 = Normalization(norm_type, groups)

        if strides == 1:
            self.downsample = None
        else:
            self.downsample = tf.keras.Sequential([
                layers.Conv2D(filters, 1, strides),
                Normalization(norm_type, groups)
            ])

    def call(self, inputs, training):
        fx = inputs
        fx = tf.nn.relu(self.norm1(self.conv1(fx), training=training))
        fx = tf.nn.relu(self.norm2(self.conv2(fx), training=training))

        if self.downsample:
            inputs = self.downsample(inputs, training=training)

        return tf.nn.relu(inputs + fx)


class BottleneckBlock(layers.Layer):
    def __init__(self, filters, norm_type='group', strides=1, **kwargs):
        super().__init__(**kwargs)
        self.filters = filters
        self.norm_type = norm_type
        self.strides = strides

        self.conv1 = layers.Conv2D(filters//4, 1)
        self.conv2 = layers.Conv2D(filters//4, 3, strides, 'same')
        self.conv3 = layers.Conv2D(filters, 1)

        groups = filters // 8
        self.norm1 = Normalization(norm_type, groups)
        self.norm2 = Normalization(norm_type, groups)
        self.norm3 = Normalization(norm_type, groups)

        if strides == 1:
            self.downsample = None
        else:
            self.downsample = tf.keras.Sequential([
                layers.Conv2D(filters, 1, strides),
                Normalization(norm_type, groups)
            ])

    def call(self, inputs, training):
        fx = inputs
        fx = tf.nn.relu(self.norm1(self.conv1(fx), training=training))
        fx = tf.nn.relu(self.norm2(self.conv2(fx), training=training))
        fx = tf.nn.relu(self.norm3(self.conv3(fx), training=training))

        if self.downsample:
            inputs = self.downsample(inputs, training=training)

        return tf.nn.relu(inputs + fx)
        

class BasicEncoder(layers.Layer):
    def __init__(self, output_dim=128, norm_type='batch', drop_rate=0.0, **kwargs):
        super().__init__(**kwargs)
        self.output_dim = output_dim
        self.norm_type = norm_type
        self.drop_rate = drop_rate

        self.conv1 = layers.Conv2D(64, 7, 2, 'same')
        self.norm1 = Normalization(norm_type, groups=8)

        self.layer1 = self._make_layer(64, strides=1)
        self.layer2 = self._make_layer(96, strides=2)
        self.layer3 = self._make_layer(128, strides=2)

        self.conv2 = layers.Conv2D(output_dim, 1)

        self.dropout = layers.Dropout(drop_rate) if drop_rate > 0 else None

    def _make_layer(self, filters, strides):
        seq = tf.keras.Sequential([
            ResBlock(filters, self.norm_type, strides),
            ResBlock(filters, self.norm_type, 1)
        ])
        return seq

    def call(self, inputs, training):
        is_list = isinstance(inputs, (tuple, list))
        if is_list:
            inputs = tf.concat(inputs, axis=0)

        x = tf.nn.relu(self.norm1(self.conv1(inputs), training=training))
        x = self.layer1(x, training=training)
        x = self.layer2(x, training=training)
        x = self.layer3(x, training=training)
        x = self.conv2(x)

        if self.dropout:
            x = self.dropout(x, training=training)

        if is_list:
            x = tf.split(x, num_or_size_splits=2, axis=0)

        return x


class SmallEncoder(layers.Layer):
    def __init__(self, output_dim=128, norm_type='batch', drop_rate=0.0, **kwargs):
        super().__init__(**kwargs)
        self.output_dim = output_dim
        self.norm_type = norm_type
        self.drop_rate = drop_rate

        self.conv1 = layers.Conv2D(32, 7, 2, 'same')
        self.norm1 = Normalization(norm_type, groups=8)

        self.layer1 = self._make_layer(32, strides=1)
        self.layer2 = self._make_layer(64, strides=2)
        self.layer3 = self._make_layer(96, strides=2)

        self.conv2 = layers.Conv2D(output_dim, 1)

        self.dropout = layers.Dropout(drop_rate) if drop_rate > 0 else None

    def _make_layer(self, filters, strides):
        seq = tf.keras.Sequential([
            ResBlock(filters, self.norm_type, strides),
            ResBlock(filters, self.norm_type, 1)
        ])
        return seq

    def call(self, inputs, training):
        is_list = isinstance(inputs, (tuple, list))
        if is_list:
            inputs = tf.concat(inputs, axis=0)
            
        x = tf.nn.relu(self.norm1(self.conv1(inputs), training=training))
        x = self.layer1(x, training=training)
        x = self.layer2(x, training=training)
        x = self.layer3(x, training=training)
        x = self.conv2(x)

        if self.dropout:
            x = self.dropout(x, training=training)

        if is_list:
            x = tf.split(x, num_or_size_splits=2, axis=0)

        return x



import tensorflow as tf


def sequence_loss(y_true, y_pred, gamma=0.8, max_flow=400):
    flow_gt, valid = y_true

    n_predictions = len(y_pred)
    flow_loss = 0.0

    # exclude invalid pixels and extremely large displacements
    mag = tf.sqrt(tf.reduce_sum(flow_gt**2, axis=-1))
    valid = valid & (mag < max_flow)
    # as float and expand channel axis
    valid = tf.expand_dims(tf.cast(valid, tf.float32), axis=-1)

    for i in range(n_predictions):
        i_weight = gamma**(n_predictions - i - 1)
        i_loss = tf.abs(y_pred[i] - flow_gt)
        flow_loss += i_weight * tf.reduce_mean(valid * i_loss)

    return flow_loss


def end_point_error(y_true, y_pred, max_flow=400):
    flow_gt, valid = y_true
        
    # exclude invalid pixels and extremely large displacements
    mag = tf.sqrt(tf.reduce_sum(flow_gt**2, axis=-1))
    valid = valid & (mag < max_flow)

    epe = tf.sqrt(tf.reduce_sum((y_pred - flow_gt)**2, axis=-1))
    epe = epe[valid]
    epe_under1 = tf.cast(epe < 1, dtype=tf.float32)
    epe_under3 = tf.cast(epe < 3, dtype=tf.float32)
    epe_under5 = tf.cast(epe < 5, dtype=tf.float32)

    result = {
        'epe': tf.reduce_mean(epe),
        'u1': tf.reduce_mean(epe_under1),
        'u3': tf.reduce_mean(epe_under3),
        'u5': tf.reduce_mean(epe_under5)
    }
    return result


class EndPointError(tf.keras.metrics.Metric):
    ''' Calculates end-point-error and relating metrics '''
    def __init__(self, max_flow=400, **kwargs):
        super().__init__(**kwargs)
        self.max_flow = max_flow

        self.epe = self.add_weight(name='epe', initializer='zeros')
        self.u1 = self.add_weight(name='u1', initializer='zeros')
        self.u3 = self.add_weight(name='u3', initializer='zeros')
        self.u5 = self.add_weight(name='u5', initializer='zeros')

        self.count = self.add_weight(name='count', initializer='zeros')

    def update_state(self, y_true, y_pred):
        flow_gt, valid = y_true
        
        # exclude invalid pixels and extremely large displacements
        mag = tf.sqrt(tf.reduce_sum(flow_gt**2, axis=-1))
        valid = valid & (mag < self.max_flow)

        epe = tf.sqrt(tf.reduce_sum((y_pred[-1] - flow_gt)**2, axis=-1))
        epe = epe[valid]
        rate_under1 = tf.cast(epe < 1, dtype=tf.float32)
        rate_under3 = tf.cast(epe < 3, dtype=tf.float32)
        rate_under5 = tf.cast(epe < 5, dtype=tf.float32)

        self.epe.assign_add(tf.reduce_mean(epe))
        self.u1.assign_add(tf.reduce_mean(rate_under1))
        self.u3.assign_add(tf.reduce_mean(rate_under3))
        self.u5.assign_add(tf.reduce_mean(rate_under5))
        self.count.assign_add(1)

    def result(self):
        count = tf.cast(self.count, dtype=self.epe.dtype)
        result = {
            'epe': self.epe / count,
            'u1': self.u1 / count,
            'u3': self.u3 / count,
            'u5': self.u5 / count
        }
        return result



import numpy as np
import cv2
cv2.setNumThreads(0)
cv2.ocl.setUseOpenCL(False)
import albumentations as A
from PIL import Image


class FlowAugmentor:
    def __init__(self, crop_size, min_scale=-0.2, max_scale=0.5, do_flip=True):
        
        # spatial augmentation params
        self.crop_size = crop_size
        self.min_scale = min_scale
        self.max_scale = max_scale
        self.spatial_aug_prob = 0.8
        self.stretch_prob = 0.8
        self.max_stretch = 0.2

        # flip augmentation params
        self.do_flip = do_flip
        self.h_flip_prob = 0.5
        self.v_flip_prob = 0.1

        # photometric augmentation params
        # self.photo_aug = ColorJitter(brightness=0.4, contrast=0.4, saturation=0.4, hue=0.5/3.14)
        self.photo_aug = A.Compose([
            A.RandomBrightnessContrast(
                brightness_limit=0.4,
                contrast_limit=0.4
            ),
            A.HueSaturationValue(
                hue_shift_limit=int(0.5/3.14*180),
                sat_shift_limit=int(0.4*255),
                val_shift_limit=int(0.)
            )
        ])
        
        self.asymmetric_color_aug_prob = 0.2
        self.eraser_aug_prob = 0.5

    def color_transform(self, img1, img2):
        """ Photometric augmentation """

        # asymmetric
        if np.random.rand() < self.asymmetric_color_aug_prob:
            # img1 = np.array(self.photo_aug(Image.fromarray(img1)), dtype=np.uint8)
            # img2 = np.array(self.photo_aug(Image.fromarray(img2)), dtype=np.uint8)
            img1 = self.photo_aug(image=img1)['image']
            img2 = self.photo_aug(image=img2)['image']

        # symmetric
        else:
            image_stack = np.concatenate([img1, img2], axis=0)
            # image_stack = np.array(self.photo_aug(Image.fromarray(image_stack)), dtype=np.uint8)
            image_stack = self.photo_aug(image=image_stack)['image']
            img1, img2 = np.split(image_stack, 2, axis=0)

        return img1, img2

    def eraser_transform(self, img1, img2, bounds=[50, 100]):
        """ Occlusion augmentation """

        ht, wd = img1.shape[:2]
        if np.random.rand() < self.eraser_aug_prob:
            mean_color = np.mean(img2.reshape(-1, 3), axis=0)
            for _ in range(np.random.randint(1, 3)):
                x0 = np.random.randint(0, wd)
                y0 = np.random.randint(0, ht)
                dx = np.random.randint(bounds[0], bounds[1])
                dy = np.random.randint(bounds[0], bounds[1])
                img2[y0:y0+dy, x0:x0+dx, :] = mean_color

        return img1, img2

    def spatial_transform(self, img1, img2, flow):
        # randomly sample scale
        ht, wd = img1.shape[:2]
        min_scale = np.maximum(
            (self.crop_size[0] + 8) / float(ht), 
            (self.crop_size[1] + 8) / float(wd))

        scale = 2 ** np.random.uniform(self.min_scale, self.max_scale)
        scale_x = scale
        scale_y = scale
        if np.random.rand() < self.stretch_prob:
            scale_x *= 2 ** np.random.uniform(-self.max_stretch, self.max_stretch)
            scale_y *= 2 ** np.random.uniform(-self.max_stretch, self.max_stretch)
        
        scale_x = np.clip(scale_x, min_scale, None)
        scale_y = np.clip(scale_y, min_scale, None)

        if np.random.rand() < self.spatial_aug_prob:
            # rescale the images
            img1 = cv2.resize(img1, None, fx=scale_x, fy=scale_y, interpolation=cv2.INTER_LINEAR)
            img2 = cv2.resize(img2, None, fx=scale_x, fy=scale_y, interpolation=cv2.INTER_LINEAR)
            flow = cv2.resize(flow, None, fx=scale_x, fy=scale_y, interpolation=cv2.INTER_LINEAR)
            flow = flow * [scale_x, scale_y]

        if self.do_flip:
            if np.random.rand() < self.h_flip_prob: # h-flip
                img1 = img1[:, ::-1]
                img2 = img2[:, ::-1]
                flow = flow[:, ::-1] * [-1.0, 1.0]

            if np.random.rand() < self.v_flip_prob: # v-flip
                img1 = img1[::-1, :]
                img2 = img2[::-1, :]
                flow = flow[::-1, :] * [1.0, -1.0]

        y0 = np.random.randint(0, img1.shape[0] - self.crop_size[0])
        x0 = np.random.randint(0, img1.shape[1] - self.crop_size[1])
        
        img1 = img1[y0:y0+self.crop_size[0], x0:x0+self.crop_size[1]]
        img2 = img2[y0:y0+self.crop_size[0], x0:x0+self.crop_size[1]]
        flow = flow[y0:y0+self.crop_size[0], x0:x0+self.crop_size[1]]

        return img1, img2, flow

    def __call__(self, img1, img2, flow):
        img1, img2 = self.color_transform(img1, img2)
        img1, img2 = self.eraser_transform(img1, img2)
        img1, img2, flow = self.spatial_transform(img1, img2, flow)

        img1 = np.ascontiguousarray(img1)
        img2 = np.ascontiguousarray(img2)
        flow = np.ascontiguousarray(flow)

        return img1, img2, flow

    
class SparseFlowAugmentor:
    def __init__(self, crop_size, min_scale=-0.2, max_scale=0.5, do_flip=False):
        # spatial augmentation params
        self.crop_size = crop_size
        self.min_scale = min_scale
        self.max_scale = max_scale
        self.spatial_aug_prob = 0.8
        self.stretch_prob = 0.8
        self.max_stretch = 0.2

        # flip augmentation params
        self.do_flip = do_flip
        self.h_flip_prob = 0.5
        self.v_flip_prob = 0.1

        # photometric augmentation params
        # self.photo_aug = ColorJitter(brightness=0.3, contrast=0.3, saturation=0.3, hue=0.3/3.14)
        self.photo_aug = A.Compose([
            A.RandomBrightnessContrast(
                brightness_limit=0.3,
                contrast_limit=0.3
            ),
            A.HueSaturationValue(
                hue_shift_limit=int(0.3/3.14*180),
                sat_shift_limit=int(0.3*255),
                val_shift_limit=int(0.)
            )
        ])        
        self.asymmetric_color_aug_prob = 0.2
        self.eraser_aug_prob = 0.5
        
    def color_transform(self, img1, img2):
        image_stack = np.concatenate([img1, img2], axis=0)
        # image_stack = np.array(self.photo_aug(Image.fromarray(image_stack)), dtype=np.uint8)
        image_stack = self.photo_aug(image=image_stack)['image']
        img1, img2 = np.split(image_stack, 2, axis=0)
        return img1, img2

    def eraser_transform(self, img1, img2):
        ht, wd = img1.shape[:2]
        if np.random.rand() < self.eraser_aug_prob:
            mean_color = np.mean(img2.reshape(-1, 3), axis=0)
            for _ in range(np.random.randint(1, 3)):
                x0 = np.random.randint(0, wd)
                y0 = np.random.randint(0, ht)
                dx = np.random.randint(50, 100)
                dy = np.random.randint(50, 100)
                img2[y0:y0+dy, x0:x0+dx, :] = mean_color

        return img1, img2

    def resize_sparse_flow_map(self, flow, valid, fx=1.0, fy=1.0):
        ht, wd = flow.shape[:2]
        coords = np.meshgrid(np.arange(wd), np.arange(ht))
        coords = np.stack(coords, axis=-1)

        coords = coords.reshape(-1, 2).astype(np.float32)
        flow = flow.reshape(-1, 2).astype(np.float32)
        valid = valid.reshape(-1).astype(np.float32)

        coords0 = coords[valid>=1]
        flow0 = flow[valid>=1]

        ht1 = int(round(ht * fy))
        wd1 = int(round(wd * fx))

        coords1 = coords0 * [fx, fy]
        flow1 = flow0 * [fx, fy]

        xx = np.round(coords1[:,0]).astype(np.int32)
        yy = np.round(coords1[:,1]).astype(np.int32)

        v = (xx > 0) & (xx < wd1) & (yy > 0) & (yy < ht1)
        xx = xx[v]
        yy = yy[v]
        flow1 = flow1[v]

        flow_img = np.zeros([ht1, wd1, 2], dtype=np.float32)
        valid_img = np.zeros([ht1, wd1], dtype=np.int32)

        flow_img[yy, xx] = flow1
        valid_img[yy, xx] = 1

        return flow_img, valid_img

    def spatial_transform(self, img1, img2, flow, valid):
        # randomly sample scale

        ht, wd = img1.shape[:2]
        min_scale = np.maximum(
            (self.crop_size[0] + 1) / float(ht), 
            (self.crop_size[1] + 1) / float(wd))

        scale = 2 ** np.random.uniform(self.min_scale, self.max_scale)
        scale_x = np.clip(scale, min_scale, None)
        scale_y = np.clip(scale, min_scale, None)

        if np.random.rand() < self.spatial_aug_prob:
            # rescale the images
            img1 = cv2.resize(img1, None, fx=scale_x, fy=scale_y, interpolation=cv2.INTER_LINEAR)
            img2 = cv2.resize(img2, None, fx=scale_x, fy=scale_y, interpolation=cv2.INTER_LINEAR)
            flow, valid = self.resize_sparse_flow_map(flow, valid, fx=scale_x, fy=scale_y)

        if self.do_flip:
            if np.random.rand() < 0.5: # h-flip
                img1 = img1[:, ::-1]
                img2 = img2[:, ::-1]
                flow = flow[:, ::-1] * [-1.0, 1.0]
                valid = valid[:, ::-1]

        margin_y = 20
        margin_x = 50

        y0 = np.random.randint(0, img1.shape[0] - self.crop_size[0] + margin_y)
        x0 = np.random.randint(-margin_x, img1.shape[1] - self.crop_size[1] + margin_x)

        y0 = np.clip(y0, 0, img1.shape[0] - self.crop_size[0])
        x0 = np.clip(x0, 0, img1.shape[1] - self.crop_size[1])

        img1 = img1[y0:y0+self.crop_size[0], x0:x0+self.crop_size[1]]
        img2 = img2[y0:y0+self.crop_size[0], x0:x0+self.crop_size[1]]
        flow = flow[y0:y0+self.crop_size[0], x0:x0+self.crop_size[1]]
        valid = valid[y0:y0+self.crop_size[0], x0:x0+self.crop_size[1]]
        return img1, img2, flow, valid

    def __call__(self, img1, img2, flow, valid):
        img1, img2 = self.color_transform(img1, img2)
        img1, img2 = self.eraser_transform(img1, img2)
        img1, img2, flow, valid = self.spatial_transform(img1, img2, flow, valid)

        img1 = np.ascontiguousarray(img1)
        img2 = np.ascontiguousarray(img2)
        flow = np.ascontiguousarray(flow)
        valid = np.ascontiguousarray(valid)

        return img1, img2, flow, valid



import numpy as np
from PIL import Image
from os.path import *
import re

import cv2
cv2.setNumThreads(0)
cv2.ocl.setUseOpenCL(False)

TAG_CHAR = np.array([202021.25], np.float32)

def readFlow(fn):
    """ Read .flo file in Middlebury format"""
    # Code adapted from:
    # http://stackoverflow.com/questions/28013200/reading-middlebury-flow-files-with-python-bytes-array-numpy

    # WARNING: this will work on little-endian architectures (eg Intel x86) only!
    # print 'fn = %s'%(fn)
    with open(fn, 'rb') as f:
        magic = np.fromfile(f, np.float32, count=1)
        if 202021.25 != magic:
            print('Magic number incorrect. Invalid .flo file')
            return None
        else:
            w = np.fromfile(f, np.int32, count=1)
            h = np.fromfile(f, np.int32, count=1)
            # print 'Reading %d x %d flo file\n' % (w, h)
            data = np.fromfile(f, np.float32, count=2*int(w)*int(h))
            # Reshape data into 3D array (columns, rows, bands)
            # The reshape here is for visualization, the original code is (w,h,2)
            return np.resize(data, (int(h), int(w), 2))

def readPFM(file):
    file = open(file, 'rb')

    color = None
    width = None
    height = None
    scale = None
    endian = None

    header = file.readline().rstrip()
    if header == b'PF':
        color = True
    elif header == b'Pf':
        color = False
    else:
        raise Exception('Not a PFM file.')

    dim_match = re.match(rb'^(\d+)\s(\d+)\s$', file.readline())
    if dim_match:
        width, height = map(int, dim_match.groups())
    else:
        raise Exception('Malformed PFM header.')

    scale = float(file.readline().rstrip())
    if scale < 0: # little-endian
        endian = '<'
        scale = -scale
    else:
        endian = '>' # big-endian

    data = np.fromfile(file, endian + 'f')
    shape = (height, width, 3) if color else (height, width)

    data = np.reshape(data, shape)
    data = np.flipud(data)
    return data

def writeFlow(filename,uv,v=None):
    """ Write optical flow to file.
    
    If v is None, uv is assumed to contain both u and v channels,
    stacked in depth.
    Original code by Deqing Sun, adapted from Daniel Scharstein.
    """
    nBands = 2

    if v is None:
        assert(uv.ndim == 3)
        assert(uv.shape[2] == 2)
        u = uv[:,:,0]
        v = uv[:,:,1]
    else:
        u = uv

    assert(u.shape == v.shape)
    height,width = u.shape
    f = open(filename,'wb')
    # write the header
    f.write(TAG_CHAR)
    np.array(width).astype(np.int32).tofile(f)
    np.array(height).astype(np.int32).tofile(f)
    # arrange into matrix form
    tmp = np.zeros((height, width*nBands))
    tmp[:,np.arange(width)*2] = u
    tmp[:,np.arange(width)*2 + 1] = v
    tmp.astype(np.float32).tofile(f)
    f.close()


def readFlowKITTI(filename):
    flow = cv2.imread(filename, cv2.IMREAD_ANYDEPTH|cv2.IMREAD_COLOR)
    flow = flow[:,:,::-1].astype(np.float32)
    flow, valid = flow[:, :, :2], flow[:, :, 2]
    flow = (flow - 2**15) / 64.0
    return flow, valid

def readDispKITTI(filename):
    disp = cv2.imread(filename, cv2.IMREAD_ANYDEPTH) / 256.0
    valid = disp > 0.0
    flow = np.stack([-disp, np.zeros_like(disp)], -1)
    return flow, valid


def writeFlowKITTI(filename, uv):
    uv = 64.0 * uv + 2**15
    valid = np.ones([uv.shape[0], uv.shape[1], 1])
    uv = np.concatenate([uv, valid], axis=-1).astype(np.uint16)
    cv2.imwrite(filename, uv[..., ::-1])
    

def read_gen(file_name, pil=False):
    ext = splitext(file_name)[-1]
    if ext == '.png' or ext == '.jpeg' or ext == '.ppm' or ext == '.jpg':
        return Image.open(file_name)
    elif ext == '.bin' or ext == '.raw':
        return np.load(file_name)
    elif ext == '.flo':
        return readFlow(file_name).astype(np.float32)
    elif ext == '.pfm':
        flow = readPFM(file_name).astype(np.float32)
        if len(flow.shape) == 2:
            return flow
        else:
            return flow[:, :, :-1]
    return []


import tensorflow as tf
from tensorflow.keras import layers
from collections import OrderedDict

from .layers.update import BasicUpdateBlock, SmallUpdateBlock
from .layers.extractor import BasicEncoder, SmallEncoder
from .layers.corr import CorrBlock, bilinear_sampler, coords_grid, upflow8


class RAFT(tf.keras.Model):
    def __init__(self, drop_rate=0, iters=12, iters_pred=24, **kwargs):
        super().__init__(**kwargs)

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128
        self.corr_levels = 4
        self.corr_radius = 4

        self.drop_rate = drop_rate

        self.iters = iters
        self.iters_pred = iters_pred

        self.fnet = BasicEncoder(output_dim=256,
                                 norm_type='instance',
                                 drop_rate=drop_rate)
        self.cnet = BasicEncoder(output_dim=hdim+cdim,
                                 norm_type='batch',
                                 drop_rate=drop_rate)
        self.update_block = BasicUpdateBlock(filters=hdim)

    def initialize_flow(self, image):
        bs, h, w, _ = image.shape
        # shape: (bs, h/8, w/8, 2)x2, grid of batch/height/width
        coords0 = coords_grid(bs, h//8, w//8)
        coords1 = coords_grid(bs, h//8, w//8)
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        ''' Upsample flow (h, w, 2) -> (8xh, 8xw, 2) using convex combination
        Args:
          flow: tensor with shape (bs, h, w, 2)
          mask: tensor with shape (bs, h, w, 64x9), 64=8x8 is the upscale
                9 is the neighborhood pixels in unfolding
        
        Returns:
          upscaled flow with shape (bs, 8xh, 8xw, 2)
        '''
        # flow: (bs, h, w, 2), mask: (bs, h, w, 64*9)
        bs, h, w, _ = flow.shape
        mask = tf.reshape(mask, (bs, h, w, 8, 8, 9, 1))
        mask = tf.nn.softmax(mask, axis=5)

        # flow: (bs, h, w, 2) -> (bs, h, w, 2*9)
        up_flow = tf.image.extract_patches(8*flow,
                                           sizes=(1, 3, 3, 1),
                                           strides=(1, 1, 1, 1),
                                           rates=(1, 1, 1, 1),
                                           padding='SAME')
        up_flow = tf.reshape(up_flow, (bs, h, w, 1, 1, 9, 2))
        # (bs, h, w, 8, 8, 9, 2) -> (bs, h, w, 8, 8, 2)
        up_flow = tf.reduce_sum(mask*up_flow, axis=5)
        # (bs, h, w, 8, 8, 2) -> (bs, h, w, 8x8x2)
        up_flow = tf.reshape(up_flow, (bs, h, w, -1))
        # (bs, h, w, 8x8x2) -> (bs, 8xh, 8xw, 2)
        return tf.nn.depth_to_space(up_flow, block_size=8)

    def call(self, inputs, training):
        image1, image2 = inputs
        image1 = 2*(image1/255.0) - 1.0
        image2 = 2*(image2/255.0) - 1.0

        # feature extractor -> (bs, h/8, w/8, 256)x2
        fmap1, fmap2 = self.fnet([image1, image2], training=training)

        # setup correlation values
        correlation = CorrBlock(fmap1, fmap2,
                                num_levels=self.corr_levels,
                                radius=self.corr_radius)

        # context network -> (bs, h/8, w/8, hdim+cdim)
        cnet = self.cnet(image1, training=training)
        # split -> (bs, h/8, w/8, hdim), (bs, h/8, w/8, cdim)
        net, inp = tf.split(cnet, [self.hidden_dim, self.context_dim], axis=-1)
        net = tf.tanh(net)
        inp = tf.nn.relu(inp)

        # (bs, h/8, w/8, 2)x2, xy-indexing
        coords0, coords1 = self.initialize_flow(image1)

        flow_predictions = []
        iters = self.iters if training else self.iters_pred
        for i in range(iters):
            # (bs, h, w, 81xnum_levels)
            corr = correlation.retrieve(coords1)

            flow = coords1 - coords0
            # (bs, h, w, *), net: hdim, up_mask: 64x9, delta_flow: 2
            net, up_mask, delta_flow = self.update_block([net, inp, corr, flow])

            # F(t+1) = F(t) + df
            coords1 += delta_flow

            # upsample prediction
            flow_up = self.upsample_flow(coords1 - coords0, up_mask)
            flow_predictions.append(flow_up)

        # flow_predictions[-1] is the finest output
        return flow_predictions

    def compile(self, optimizer, clip_norm, loss, epe, **kwargs):
        super().compile(**kwargs)
        self.optimizer = optimizer
        self.clip_norm = clip_norm
        self.loss = loss
        self.epe = epe

        self.flow_metrics = OrderedDict({
            'loss': tf.keras.metrics.Mean(name='loss'),
            'epe': tf.keras.metrics.Mean(name='epe'),
            'u1': tf.keras.metrics.Mean(name='u1'),
            'u3': tf.keras.metrics.Mean(name='u3'),
            'u5': tf.keras.metrics.Mean(name='u5')
        })

    def train_step(self, data):
        image1, image2, flow, valid = data
        image1 = tf.cast(image1, dtype=tf.float32)
        image2 = tf.cast(image2, dtype=tf.float32)

        with tf.GradientTape() as tape:
            flow_predictions = self([image1, image2], training=True)
            loss = self.loss([flow, valid], flow_predictions)
        grads = tape.gradient(loss, self.trainable_weights)
        grads, _ = tf.clip_by_global_norm(grads, self.clip_norm)
        self.optimizer.apply_gradients(zip(grads, self.trainable_weights))

        info = self.epe([flow, valid], flow_predictions[-1])
        self.flow_metrics['loss'].update_state(loss)
        self.flow_metrics['epe'].update_state(info['epe'])
        self.flow_metrics['u1'].update_state(info['u1'])
        self.flow_metrics['u3'].update_state(info['u3'])
        self.flow_metrics['u5'].update_state(info['u5'])
        return {k: m.result() for k, m in self.flow_metrics.items()}

    def test_step(self, data):
        image1, image2, flow, valid = data
        image1 = tf.cast(image1, dtype=tf.float32)
        image2 = tf.cast(image2, dtype=tf.float32)

        flow_predictions = self([image1, image2], training=False)

        info = self.epe([flow, valid], flow_predictions[-1])
        self.flow_metrics['epe'].update_state(info['epe'])
        self.flow_metrics['u1'].update_state(info['u1'])
        self.flow_metrics['u3'].update_state(info['u3'])
        self.flow_metrics['u5'].update_state(info['u5'])
        return {k: m.result() for k, m in self.flow_metrics.items()}

    def predict_step(self, data):
        image1, image2, *_ = data
        image1 = tf.cast(image1, dtype=tf.float32)
        image2 = tf.cast(image2, dtype=tf.float32)

        flow_predictions = self([image1, image2], training=False)
        return flow_predictions[-1]

    def reset_metrics(self):
        for k, m in self.flow_metrics.items():
            m.reset_states()


class SmallRAFT(RAFT):
    def __init__(self, drop_rate=0, iters=12, iters_pred=24, **kwargs):
        super().__init__(drop_rate, iters, iters_pred, **kwargs)

        self.hidden_dim = hdim = 96
        self.context_dim = cdim = 64
        self.corr_levels = 4
        self.corr_radius = 3

        self.fnet = SmallEncoder(output_dim=128,
                                 norm_type='instance',
                                 drop_rate=drop_rate)
        self.cnet = SmallEncoder(output_dim=hdim+cdim,
                                 norm_type=None,
                                 drop_rate=drop_rate)
        self.update_block = SmallUpdateBlock(filters=hdim)

    def call(self, inputs, training):
        image1, image2 = inputs
        image1 = 2*(image1/255.0) - 1.0
        image2 = 2*(image2/255.0) - 1.0

        # feature extractor -> (bs, h/8, w/8, nch)x2
        fmap1, fmap2 = self.fnet([image1, image2], training=training)

        # setup correlation values
        correlation = CorrBlock(fmap1, fmap2,
                                num_levels=self.corr_levels,
                                radius=self.corr_radius)

        # context network
        cnet = self.cnet(image1, training=training)
        net, inp = tf.split(cnet, [self.hidden_dim, self.context_dim], axis=-1)
        net = tf.tanh(net)
        inp = tf.nn.relu(inp)

        coords0, coords1 = self.initialize_flow(image1)

        flow_predictions = []
        iters = self.iters if training else self.iters_pred
        for i in range(iters):
            corr = correlation.retrieve(coords1)

            flow = coords1 - coords0
            net, _, delta_flow = self.update_block([net, inp, corr, flow])

            # F(t+1) = F(t) + df
            coords1 += delta_flow

            # upsample prediction
            flow_up = upflow8(coords1 - coords0)
            flow_predictions.append(flow_up)

        return flow_predictions        



import tensorflow as tf
from tensorflow.keras import layers


class FlowHead(layers.Layer):
    def __init__(self, filters=256, **kwargs):
        super().__init__(**kwargs)
        self.filters = filters

        self.conv1 = layers.Conv2D(filters, 3, 1, 'same')
        self.conv2 = layers.Conv2D(2, 3, 1, 'same')

    def call(self, inputs):
        return self.conv2(tf.nn.relu(self.conv1(inputs)))
        

class ConvGRU(layers.Layer):
    def __init__(self, filters=128, **kwargs):
        super().__init__(**kwargs)
        self.filters = filters

        self.convz = layers.Conv2D(filters, 3, 1, 'same')
        self.convr = layers.Conv2D(filters, 3, 1, 'same')
        self.convq = layers.Conv2D(filters, 3, 1, 'same')
    
    def call(self, inputs):
        h, x = inputs
        hx = tf.concat([h, x], axis=-1)

        z = tf.nn.sigmoid(self.convz(hx))
        r = tf.nn.sigmoid(self.convr(hx))
        q = tf.nn.tanh(self.convq(tf.concat([r*h, x], axis=-1)))

        h = (1-z)*h + z*q
        return h


class SepConvGRU(layers.Layer):
    def __init__(self, filters=128, **kwargs):
        super().__init__(**kwargs)
        self.filters = filters

        self.convz1 = layers.Conv2D(filters, (1, 5), 1, 'same')
        self.convr1 = layers.Conv2D(filters, (1, 5), 1, 'same')
        self.convq1 = layers.Conv2D(filters, (1, 5), 1, 'same')

        self.convz2 = layers.Conv2D(filters, (5, 1), 1, 'same')
        self.convr2 = layers.Conv2D(filters, (5, 1), 1, 'same')
        self.convq2 = layers.Conv2D(filters, (5, 1), 1, 'same')

    def call(self, inputs):
        h, x = inputs
        hx = tf.concat([h, x], axis=-1)
        # horizontal
        z = tf.nn.sigmoid(self.convz1(hx))
        r = tf.nn.sigmoid(self.convr1(hx))
        q = tf.nn.tanh(self.convq1(tf.concat([r*h, x], axis=-1)))
        h = (1-z)*h + z*q

        # vertical
        hx = tf.concat([h, x], axis=-1)
        z = tf.nn.sigmoid(self.convz2(hx))
        r = tf.nn.sigmoid(self.convr2(hx))
        q = tf.nn.tanh(self.convq2(tf.concat([r*h, x], axis=-1)))
        h = (1-z)*h + z*q
        
        return h
    

class SmallMotionEncoder(layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.convc1 = layers.Conv2D(96, 1, 1, 'same')
        self.convf1 = layers.Conv2D(64, 7, 1, 'same')
        self.convf2 = layers.Conv2D(32, 3, 1, 'same')
        self.conv = layers.Conv2D(80, 3, 1, 'same')

    def call(self, inputs):
        flow, corr = inputs
        cor = tf.nn.relu(self.convc1(corr))
        flo = tf.nn.relu(self.convf1(flow))
        flo = tf.nn.relu(self.convf2(flo))
        cor_flo = tf.concat([cor, flo], axis=-1)
        out = tf.nn.relu(self.conv(cor_flo))
        return tf.concat([out, flow], axis=-1)


class BasicMotionEncoder(layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.convc1 = layers.Conv2D(256, 1, 1)
        self.convc2 = layers.Conv2D(192, 3, 1, 'same')
        self.convf1 = layers.Conv2D(128, 7, 1, 'same')
        self.convf2 = layers.Conv2D(64, 3, 1, 'same')
        self.conv = layers.Conv2D(128-2, 3, 1, 'same')

    def call(self, inputs):
        flow, corr = inputs
        cor = tf.nn.relu(self.convc1(corr))
        cor = tf.nn.relu(self.convc2(cor))
        flo = tf.nn.relu(self.convf1(flow))
        flo = tf.nn.relu(self.convf2(flo))

        cor_flo = tf.concat([cor, flo], axis=-1)
        out = tf.nn.relu(self.conv(cor_flo))
        return tf.concat([out, flow], axis=-1)


class SmallUpdateBlock(layers.Layer):
    def __init__(self, filters=96, **kwargs):
        super().__init__(**kwargs)
        self.filters = filters

        self.encoder = SmallMotionEncoder()
        self.gru = ConvGRU(filters)
        self.flow_head = FlowHead(128)

    def call(self, inputs):
        net, inp, corr, flow = inputs
        motion_features = self.encoder([flow, corr])
        inp = tf.concat([inp, motion_features], axis=-1)
        net = self.gru([net, inp])
        delta_flow = self.flow_head(net)

        return net, None, delta_flow


class BasicUpdateBlock(layers.Layer):
    def __init__(self, filters=128, **kwargs):
        super().__init__(**kwargs)
        self.filters = filters

        self.encoder = BasicMotionEncoder()
        self.gru = SepConvGRU(filters)
        self.flow_head = FlowHead(256)

        self.mask = tf.keras.Sequential([
            layers.Conv2D(256, 3, 1, 'same'),
            layers.ReLU(),
            layers.Conv2D(64*9, 1, 1)
        ])

    def call(self, inputs):
        net, inp, corr, flow = inputs
        motion_features = self.encoder([flow, corr])
        inp = tf.concat([inp, motion_features], axis=-1)

        net = self.gru([net, inp])
        delta_flow = self.flow_head(net)

        # scale mask to balance gradients
        mask = 0.25*self.mask(net)
        return net, mask, delta_flow
    



import os
import imageio
import numpy as np
import tensorflow as tf
from tensorflow.keras import callbacks

from .datasets.flow_viz import flow_to_image


def first_cycle_scaler(cycle):
    ''' Return learning rate scale 
    |--    cycle = 1    --|-- cycle > 1 --
    min_lr -> max_lr -> min_lr -> (const)
    '''
    return tf.cond(cycle == 1, lambda: 1.0, lambda: 0.0)


def inverse_scaler(cycle):
    ''' Return learning rate scale
    |--    cycle = 1    --|--    cycle = 2    --|--   cycle = 3 --
    min_lr -> max_lr -> min_lr -> max_lr/2 -> min_lr -> max_lr/3 -> ...
    '''
    return 1/cycle


class VisFlowCallback(callbacks.Callback):
    ''' Class for visualization of predicted flow '''
    def __init__(self,
                 dataset,
                 target_size=(448, 1024),
                 num_visualize=1,
                 choose_random=False,
                 logdir='predicted_flows',
                 **kwargs):
        '''
        Args:
          dataset (FlowDataset): iterating input images and extra info
            (without GT-flow)
          target_size (tuple of int): size for properly processed in the model.
            RAFT requires size multiple of 64. Default is (448, 1024).
          num_visualize (int): number of images to be saved. Default is 1
          choose_random (bool): whether to shuffle target image. Default is False.
          logdir (str): log directory for resulting visualization. Default is 'logs'
        '''
        super().__init__(**kwargs)
        self.dataset = dataset
        self.target_size = target_size
        self.num_visualize = num_visualize
        self.choose_random = choose_random
        self.logdir = logdir

        if not os.path.exists(logdir):
            os.makedirs(logdir)

    def on_epoch_end(self, epoch, logs=None):
        if self.choose_random:
            vis_ids = np.random.choice(len(self.dataset),
                                       size=self.num_visualize,
                                       replace=False)
        else:
            vis_ids = range(self.num_visualize)

        for i in vis_ids:
            image1, image2, *_ = self.dataset[i]
            if len(image1.shape) > 3:
                raise ValueError('target dataset must not be batched')

            h_origin, w_origin, _ = image1.shape

            image1_tf = tf.convert_to_tensor(image1, dtype=tf.float32)
            image2_tf = tf.convert_to_tensor(image2, dtype=tf.float32)

            image1_tf = tf.image.resize_with_crop_or_pad(image1_tf, *self.target_size)
            image2_tf = tf.image.resize_with_crop_or_pad(image2_tf, *self.target_size)

            image1_tf = tf.expand_dims(image1_tf, axis=0)
            image2_tf = tf.expand_dims(image2_tf, axis=0)

            flow_predictions = self.model([image1_tf, image2_tf], training=False)
            flow_pred = flow_predictions[-1][0]
            flow_pred = tf.image.resize_with_crop_or_pad(flow_pred, h_origin, w_origin)
            flow_img = flow_to_image(flow_pred.numpy()).astype(np.uint8)

            contents = np.concatenate([image1, image2, flow_img], axis=0)

            filename = f'epoch{str(epoch+1).zfill(3)}_{str(i+1).zfill(3)}.png'
            savepath = os.path.join(self.logdir, filename)
            imageio.imwrite(savepath, contents)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tf_raft', 'tf_raft.datasets', 'tf_raft.layers', 'tf_raft.losses']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=0.4.6,<0.5.0',
 'tensorflow-addons>=0.11.1,<0.12.0',
 'tensorflow>=2.3.0,<3.0.0']

setup_kwargs = {
    'name': 'tf-raft',
    'version': '0.1.4',
    'description': 'RAFT (Recurrent All Pairs Field Transforms for Optical Flow) implementation via tf.keras',
    'long_description': "# tf-raft\nRAFT (Recurrent All Pairs Field Transforms for Optical Flow, Teed et. al., ECCV2020) implementation via tf.keras\n\n## Original resources\n- [RAFT: Recurrent All Pairs Field Transforms for Optical Flow](https://arxiv.org/abs/2003.12039)\n- https://github.com/princeton-vl/RAFT\n\n## Installation\n\n```\n$ pip install tf-raft\n```\n\nor you can simply clone this repository.\n\n### Dependencies\n- TensorFlow\n- TensorFlow-addons\n- albumentations\n\nsee details in `pyoroject.toml`\n\n## Optical flow datasets\n[MPI-Sintel](http://sintel.is.tue.mpg.de/) or [FlyingChairs](https://lmb.informatik.uni-freiburg.de/resources/datasets/FlyingChairs.en.html#flyingchairs) datasets are relatively light. See more datasets in the [oirignal repository](https://github.com/princeton-vl/RAFT)\n\n## Usage\n\n``` python\nfrom tf_raft.model import RAFT, SmallRAFT\nfrom tf_raft.losses import sequence_loss, end_point_error\n\n# iters/iters_pred are the number of recurrent update of flow in training/prediction\nraft = RAFT(iters=iters, iters_pred=iters_pred)\nraft.compile(\n    optimizer=optimizer,\n    clip_norm=clip_norm,\n    loss=sequence_loss,\n    epe=end_point_error\n)\n\nraft.fit(\n    ds_train,\n    epochs=epochs,\n    callbacks=callbacks,\n    steps_per_epoch=train_size//batch_size,\n    validation_data=ds_val,\n    validation_steps=val_size\n)\n```\n\nIn practice, you are required to prepare dataset, optimizer, callbacks etc, check details in `train_sintel.py` or `train_chairs.py`.\n\n### Train via YAML configuration\n\n`train_chairs.py` and `train_sintel.py` train RAFT model via YAML configuration. Sample configs are in `configs` directory. Run;\n\n``` shell\n$ python train_chairs.py /path/to/config.yml\n```\n\n## Pre-trained models\n\nI made the pre-trained weights (on both FlyingChairs and MPI-Sintel) public.\nYou can download them via `gsutil` or `curl`.\n\n### Trained weights on FlyingChairs\n\n``` shell\n$ gsutil cp -r gs://tf-raft-pretrained/2020-09-26T18-38/checkpoints .\n```\nor\n``` shell\n$ mkdir checkpoints\n$ curl -OL https://storage.googleapis.com/tf-raft-pretrained/2020-09-26T18-38/checkpoints/model.data-00000-of-00001\n$ curl -OL https://storage.googleapis.com/tf-raft-pretrained/2020-09-26T18-38/checkpoints/model.index\n$ mv model* checkpoints/\n```\n\n### Trained weights on MPI-Sintel (Clean path)\n\n``` shell\n$ gsutil cp -r gs://tf-raft-pretrained/2020-09-26T08-51/checkpoints .\n```\nor\n``` shell\n$ mkdir checkpoints\n$ curl -OL https://storage.googleapis.com/tf-raft-pretrained/2020-09-26T08-51/checkpoints/model.data-00000-of-00001\n$ curl -OL https://storage.googleapis.com/tf-raft-pretrained/2020-09-26T08-51/checkpoints/model.index\n$ mv model* checkpoints/\n```\n\n### Load weights\n\n``` python\nraft = RAFT(iters=iters, iters_pred=iters_pred)\nraft.load_weights('checkpoints/model')\n\n# forward (with dummy inputs)\nx1 = np.random.uniform(0, 255, (1, 448, 512, 3)).astype(np.float32)\nx2 = np.random.uniform(0, 255, (1, 448, 512, 3)).astype(np.float32)\nflow_predictions = model([x1, x2], training=False)\n\nprint(flow_predictions[-1].shape) # >> (1, 448, 512, 2)\n```\n\n## Note\nThough I have tried to reproduce the original implementation faithfully, there is some difference between the original one and mine (mainly because of used framework: PyTorch/TensorFlow);\n\n- The original implementations provides cuda-based correlation function but I don't. My TF-based implementation works well, but cuda-based one may run faster.\n- I have trained my model on FlyingChairs and MPI-Sintel separately in my private environment (GCP with P100 accelerator). The model has been trained well, but not reached the best score reported in the paper (trained on multiple datasets).\n- The original one uses mixed-precision. This may get training much faster, but I don't. TensorFlow also enables mixed-precision with few additional lines, see https://www.tensorflow.org/guide/mixed_precision if interested.\n\nAdditional, global gradient clipping seems to be essential for stable training though it is not emphasized in the original paper. This operation can be done via `torch.nn.utils.clip_grad_norm_(model.parameters(), clip)` in PyTorch, `tf.clip_by_global_norm(grads, clip_norm)` in TF (coded at `self.train_step` in `tf_raft/model.py`).\n\n## References\n- https://github.com/princeton-vl/RAFT\n- https://github.com/NVIDIA/flownet2-pytorch\n- https://github.com/NVlabs/PWC-Net\n",
    'author': 'Daigo Hirooka',
    'author_email': 'daigo.hirooka@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/daigo0927/tf-raft',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Copied and edited from https://github.com/princeton-vl/RAFT/blob/master/core/datasets.py
# which is under BSD 3-Clause License

# Data loading based on https://github.com/NVIDIA/flownet2-pytorch

import numpy as np
import tensorflow as tf

import os
import math
import copy
import random
from glob import glob
import os.path as osp

from . import frame_utils
from .augmentor import FlowAugmentor, SparseFlowAugmentor


class FlowDataset:
    ''' Base class for optical flow dataset iterations '''
    def __init__(self, aug_params=None, sparse=False):
        ''' 
        Args:
          aug_params (dict): A dict containing augmentation parameters
            treated as arguments for FlowAugmentor.
            - crop_size (tuple<int>): Spartial crop size
            - min_scale (float): minimum scale factor
            - max_scale (float): maximum scale factor
            - do_flip (bool): flip flag
          sparse (bool): Whether the flow data is annotated sparse or not 
        '''
        self.augmentor = None
        self.sparse = sparse
        if aug_params is not None:
            if sparse:
                self.augmentor = SparseFlowAugmentor(**aug_params)
            else:
                self.augmentor = FlowAugmentor(**aug_params)

        self.is_test = False
        self.init_seed = False
        self.flow_list = []
        self.image_list = []
        self.extra_info = []

    def __getitem__(self, index):
        if self.is_test:
            img1 = frame_utils.read_gen(self.image_list[index][0])
            img2 = frame_utils.read_gen(self.image_list[index][1])
            img1 = np.array(img1).astype(np.uint8)[..., :3]
            img2 = np.array(img2).astype(np.uint8)[..., :3]
            # img1 = torch.from_numpy(img1).permute(2, 0, 1).float()
            # img2 = torch.from_numpy(img2).permute(2, 0, 1).float()
            return img1, img2, self.extra_info[index]

        # if not self.init_seed:
        #     # worker_info = torch.utils.data.get_worker_info()
        #     if worker_info is not None:
        #         # torch.manual_seed(worker_info.id)
        #         np.random.seed(worker_info.id)
        #         random.seed(worker_info.id)
        #         self.init_seed = True

        index = index % len(self.image_list)
        valid = None
        if self.sparse:
            flow, valid = frame_utils.readFlowKITTI(self.flow_list[index])
        else:
            flow = frame_utils.read_gen(self.flow_list[index])

        img1 = frame_utils.read_gen(self.image_list[index][0])
        img2 = frame_utils.read_gen(self.image_list[index][1])

        flow = np.array(flow).astype(np.float32)
        img1 = np.array(img1).astype(np.uint8)
        img2 = np.array(img2).astype(np.uint8)

        # grayscale images
        if len(img1.shape) == 2:
            img1 = np.tile(img1[...,None], (1, 1, 3))
            img2 = np.tile(img2[...,None], (1, 1, 3))
        else:
            img1 = img1[..., :3]
            img2 = img2[..., :3]

        if self.augmentor is not None:
            if self.sparse:
                img1, img2, flow, valid = self.augmentor(img1, img2, flow, valid)
            else:
                img1, img2, flow = self.augmentor(img1, img2, flow)

        # img1 = torch.from_numpy(img1).permute(2, 0, 1).float()
        # img2 = torch.from_numpy(img2).permute(2, 0, 1).float()
        # flow = torch.from_numpy(flow).permute(2, 0, 1).float()

        if valid is not None:
            # valid = torch.from_numpy(valid)
            pass
        else:
            # valid = (flow[:, :, 0].abs() < 1000) & (flow[:, :, 1].abs() < 1000)
            valid = (np.abs(flow[:, :, 0]) < 1000) * (np.abs(flow[:, :, 1]) < 1000)

        # return img1, img2, flow, valid.float()
        return img1, img2, flow, valid

    def __rmul__(self, v):
        self_copy = copy.deepcopy(self)
        self_copy.flow_list *= v
        self_copy.image_list *= v
        return self_copy

    def __add__(self, other):
        copied = copy.deepcopy(self)
        copied.flow_list += other.flow_list
        copied.image_list += other.image_list
        return copied
        
    def __len__(self):
        return len(self.image_list)

    def __call__(self):
        for sample in self:
            yield sample

    def shuffle(self):
        perm = np.random.permutation(len(self))
        self.flow_list = [self.flow_list[i] for i in perm]
        self.image_list = [self.image_list[i] for i in perm]
        

class MpiSintel(FlowDataset):
    '''MPI Sintel dataset: http://sintel.is.tue.mpg.de/ '''
    def __init__(self,
                 aug_params=None,
                 split='training',
                 root='datasets/MPI-Sintel-complete',
                 dstype='clean'):
        ''' 
        Args:
          aug_params (dict): A dict containing augmentation parameters
            treated as arguments for FlowAugmentor.
            - crop_size (tuple<int>): Spartial crop size
            - min_scale (float): minimum scale factor
            - max_scale (float): maximum scale factor
            - do_flip (bool): flip flag
          split (str): training/validation split
          root (str): path to the dataset directory
          dstype (str): clean/final path (difficulty)
        '''        
        super(MpiSintel, self).__init__(aug_params)
        flow_root = osp.join(root, split, 'flow')
        image_root = osp.join(root, split, dstype)

        if split == 'test':
            self.is_test = True

        for scene in os.listdir(image_root):
            image_list = sorted(glob(osp.join(image_root, scene, '*.png')))
            for i in range(len(image_list)-1):
                self.image_list += [ [image_list[i], image_list[i+1]] ]
                self.extra_info += [ (scene, i) ] # scene and frame_id

            if split != 'test':
                self.flow_list += sorted(glob(osp.join(flow_root, scene, '*.flo')))


class FlyingChairs(FlowDataset):
    ''' FlyingChairs dataset:
     https://lmb.informatik.uni-freiburg.de/resources/datasets/FlyingChairs.en.html
     '''
    def __init__(self,
                 aug_params=None,
                 split='training',
                 split_txt='FlyingChairs_train_val.txt',
                 root='datasets/FlyingChairs_release/data'):
        ''' 
        Args:
          aug_params (dict): A dict containing augmentation parameters
            treated as arguments for FlowAugmentor.
            - crop_size (tuple<int>): Spartial crop size
            - min_scale (float): minimum scale factor
            - max_scale (float): maximum scale factor
            - do_flip (bool): flip flag
          split (str): training/validation split
          split_txt (str): path to the textfile indicating train/val split
          root (str): path to the dataset directory
        '''        
        super(FlyingChairs, self).__init__(aug_params)

        images = sorted(glob(osp.join(root, '*.ppm')))
        flows = sorted(glob(osp.join(root, '*.flo')))
        assert (len(images)//2 == len(flows))

        split_list = np.loadtxt(split_txt, dtype=np.int32)
        for i in range(len(flows)):
            xid = split_list[i]
            if (split=='training' and xid==1) or (split=='validation' and xid==2):
                self.flow_list += [ flows[i] ]
                self.image_list += [ [images[2*i], images[2*i+1]] ]


class FlyingThings3D(FlowDataset):
    def __init__(self,
                 aug_params=None,
                 root='datasets/FlyingThings3D',
                 dstype='frames_cleanpass'):
        super(FlyingThings3D, self).__init__(aug_params)

        for cam in ['left']:
            for direction in ['into_future', 'into_past']:
                image_dirs = sorted(glob(osp.join(root, dstype, 'TRAIN/*/*')))
                image_dirs = sorted([osp.join(f, cam) for f in image_dirs])

                flow_dirs = sorted(glob(osp.join(root, 'optical_flow/TRAIN/*/*')))
                flow_dirs = sorted([osp.join(f, direction, cam) for f in flow_dirs])

                for idir, fdir in zip(image_dirs, flow_dirs):
                    images = sorted(glob(osp.join(idir, '*.png')) )
                    flows = sorted(glob(osp.join(fdir, '*.pfm')) )
                    for i in range(len(flows)-1):
                        if direction == 'into_future':
                            self.image_list += [ [images[i], images[i+1]] ]
                            self.flow_list += [ flows[i] ]
                        elif direction == 'into_past':
                            self.image_list += [ [images[i+1], images[i]] ]
                            self.flow_list += [ flows[i+1] ]
      

class KITTI(FlowDataset):
    def __init__(self,
                 aug_params=None,
                 split='training',
                 root='datasets/KITTI'):
        super(KITTI, self).__init__(aug_params, sparse=True)
        if split == 'testing':
            self.is_test = True

        root = osp.join(root, split)
        images1 = sorted(glob(osp.join(root, 'image_2/*_10.png')))
        images2 = sorted(glob(osp.join(root, 'image_2/*_11.png')))

        for img1, img2 in zip(images1, images2):
            frame_id = img1.split('/')[-1]
            self.extra_info += [ [frame_id] ]
            self.image_list += [ [img1, img2] ]

        if split == 'training':
            self.flow_list = sorted(glob(osp.join(root, 'flow_occ/*_10.png')))


class HD1K(FlowDataset):
    def __init__(self, aug_params=None, root='datasets/HD1k'):
        super(HD1K, self).__init__(aug_params, sparse=True)

        seq_ix = 0
        while 1:
            flows = sorted(glob(os.path.join(root, 'hd1k_flow_gt', 'flow_occ/%06d_*.png' % seq_ix)))
            images = sorted(glob(os.path.join(root, 'hd1k_input', 'image_2/%06d_*.png' % seq_ix)))

            if len(flows) == 0:
                break

            for i in range(len(flows)-1):
                self.flow_list += [flows[i]]
                self.image_list += [ [images[i], images[i+1]] ]

            seq_ix += 1


def fetch_dataloader(args, TRAIN_DS='C+T+K+S+H'):
    """ Create the data loader for the corresponding trainign set """

    if args.stage == 'chairs':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.1, 'max_scale': 1.0, 'do_flip': True}
        train_dataset = FlyingChairs(aug_params, split='training')
    
    elif args.stage == 'things':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.4, 'max_scale': 0.8, 'do_flip': True}
        clean_dataset = FlyingThings3D(aug_params, dstype='frames_cleanpass')
        final_dataset = FlyingThings3D(aug_params, dstype='frames_finalpass')
        train_dataset = clean_dataset + final_dataset

    elif args.stage == 'sintel':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.2, 'max_scale': 0.6, 'do_flip': True}
        things = FlyingThings3D(aug_params, dstype='frames_cleanpass')
        sintel_clean = MpiSintel(aug_params, split='training', dstype='clean')
        sintel_final = MpiSintel(aug_params, split='training', dstype='final')        

        if TRAIN_DS == 'C+T+K+S+H':
            kitti = KITTI({'crop_size': args.image_size, 'min_scale': -0.3, 'max_scale': 0.5, 'do_flip': True})
            hd1k = HD1K({'crop_size': args.image_size, 'min_scale': -0.5, 'max_scale': 0.2, 'do_flip': True})
            train_dataset = 100*sintel_clean + 100*sintel_final + 200*kitti + 5*hd1k + things

        elif TRAIN_DS == 'C+T+K/S':
            train_dataset = 100*sintel_clean + 100*sintel_final + things

    elif args.stage == 'kitti':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.2, 'max_scale': 0.4, 'do_flip': False}
        train_dataset = KITTI(aug_params, split='training')

    train_loader = data.DataLoader(train_dataset, batch_size=args.batch_size, 
        pin_memory=False, shuffle=True, num_workers=4, drop_last=True)

    print('Training with %d image pairs' % len(train_dataset))
    return train_loader


def ShapeSetter(batch_size, image_size):
    def f(image1, image2, flow, valid):
        image1.set_shape((batch_size, *image_size, 3))
        image2.set_shape((batch_size, *image_size, 3))
        flow.set_shape((batch_size, *image_size, 2))
        valid.set_shape((batch_size, *image_size))
        return image1, image2, flow, valid
    return f
    

def as_supervised(image1, image2, flow, valid):
    return (image1, image2), (flow, valid)


def CropOrPadder(target_size):
    def f(image1, image2, flow, valid):
        image1 = tf.image.resize_with_crop_or_pad(image1, *target_size)
        image2 = tf.image.resize_with_crop_or_pad(image2, *target_size)
        
        flow = tf.image.resize_with_crop_or_pad(flow, *target_size)
        valid = tf.expand_dims(valid, axis=-1)
        valid = tf.image.resize_with_crop_or_pad(valid, *target_size)
        valid = tf.squeeze(valid, axis=-1)        

        return image1, image2, flow, valid
    return f



import tensorflow as tf
import tensorflow_addons as tfa
from tensorflow.keras import layers


def tfa_sampler(image, coords, mask=False):
    ''' Value sampler using tfa.image.resampler
    Args:
      image: tensor with shape (bs*h*w, h/i**2, w/i**2, 1)
      coords: coordinates tensor with shape (bs*h*w,, 2r+1, 2r+1, 2), xy-indexing
      mask: not implemented (same as the original implementation)
      
    Returns:
      sampled tensor with shape (bs*h*w, 2r+1, 2r+1, 1)

    None that tfa.image.resampler is decorated with @tf.function.
    Recursive use of this can be expensive and exessive.
    '''
    _, h, w, _ = image.shape
    # output: (bs*h*w, 2r+1, 2r+1, 1)
    image = tfa.image.resampler(image, coords)

    if mask:
        raise NotImplementedError("mask is not implemented for True")
    return image


def bilinear_sampler(image, coords):
    ''' Value sampler using tf.gather_nd
    Args:
      image: tensor with shape (bs*h*w, h/i**2, w/i**2, 1)
      coords: coordinates tensor with shape (bs*h*w,, 2r+1, 2r+1, 2), xy-indexing
      mask: not implemented (same as the original implementation)
      
    Returns:
      sampled tensor with shape (bs*h*w, 2r+1, 2r+1, 1)
    '''    
    _, h, w, _ = image.shape
    # -> (bs*h*w, 2r+1, 2r+1)x2
    gx, gy = tf.unstack(coords, axis=-1)
    gx = tf.clip_by_value(gx, 0, w-1)
    gy = tf.clip_by_value(gy, 0, h-1)

    # corners: (bs*h*w, 2r+1, 2r+1)x4
    gx0 = tf.floor(gx)
    gx1 = tf.math.ceil(gx)
    gy0 = tf.floor(gy)
    gy1 = tf.math.ceil(gy)

    # coordinates: (bs*h*w, 2r+1, 2r+1, 2)x4
    g00 = tf.stack([gy0, gx0], axis=-1)
    g01 = tf.stack([gy0, gx1], axis=-1)
    g10 = tf.stack([gy1, gx0], axis=-1)
    g11 = tf.stack([gy1, gx1], axis=-1)

    # coefficients: (bs*h*w, 2r+1, 2r+1, 1)x4
    c00 = tf.expand_dims((gy1 - gy)*(gx1 - gx), axis=-1)
    c01 = tf.expand_dims((gy1 - gy)*(gx - gx0), axis=-1)
    c10 = tf.expand_dims((gy - gy0)*(gx1 - gx), axis=-1)
    c11 = tf.expand_dims((gy - gy0)*(gx - gx0), axis=-1)

    # gathered values: (bs*h*w, 2r+1, 2r+1, 1)
    x00 = tf.gather_nd(image, tf.cast(g00, dtype=tf.int32), batch_dims=1)
    x01 = tf.gather_nd(image, tf.cast(g01, dtype=tf.int32), batch_dims=1)
    x10 = tf.gather_nd(image, tf.cast(g10, dtype=tf.int32), batch_dims=1)
    x11 = tf.gather_nd(image, tf.cast(g11, dtype=tf.int32), batch_dims=1)

    output = c00 * x00 + c01 * x01 + c10 * x10 + c11 * x11
    return output


def coords_grid(batch_size, height, width):
    ''' Generate coordinates (xy-order) from given info
    Args:
      batch_size, height, width: int values

    Returns:
      coordinates tensor with shape (bs, h, w, 2), xy-indexing
    '''
    # shape: (height, width)x2
    gy, gx = tf.meshgrid(tf.range(height, dtype=tf.float32),
                         tf.range(width, dtype=tf.float32),
                         indexing='ij')
    # -> (height, width, 2)
    coords = tf.stack([gx, gy], axis=-1)
    # -> (1, height, width, 2)
    coords = tf.expand_dims(coords, axis=0)
    # -> (batch_size, height, width, 2)
    coords = tf.tile(coords, (batch_size, 1, 1, 1))
    return coords


def upflow8(flow, mode='bilinear'):
    _, h, w, _ = flow.shape
    new_size = (8*h, 8*w)
    return 8*tf.image.resize(flow, new_size, mode)


class CorrBlock:
    def __init__(self, fmap1, fmap2, num_levels=4, radius=4):
        self.fmap1 = fmap1
        self.fmap2 = fmap2
        self.num_levels = num_levels
        self.radius = radius

        corr = self.correlation(fmap1, fmap2)
        batch_size, h1, w1, _, h2, w2 = corr.shape
        corr = tf.reshape(corr, (batch_size*h1*w1, h2, w2, 1))

        # (bs*h*w, h, w, 1), (bs*h*w, h/2, w/2, 1), ..., (bs*h*w, h/8, w/8, 1)
        self.corr_pyramid = [corr] 
        for _ in range(num_levels - 1):
            corr = tf.nn.avg_pool2d(corr, 2, 2, padding='VALID')
            self.corr_pyramid.append(corr)

    def retrieve(self, coords):
        ''' Retrieve correlation values specified by coordinates
        Args:
          coords: coordinates tensor, shape (batch_size, h, w, 2)
        
        Returns:
          A tensor contains multiscale correlation
            with shape (bs, h, w, 81*num_levels)
        '''
        r = self.radius
        bs, h, w, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            # (bs*h*w, h, w, 1), (bs*h*w, h/2, w/2, 1), ..., (bs*h*w, h/8, w/8, 1)
            corr = self.corr_pyramid[i]
            # (2r+1, 2r+1)x2
            d = tf.range(-r, r+1, dtype=tf.float32)
            dy, dx = tf.meshgrid(d, d, indexing='ij')
            # (2r+1, 2r+1, 2)
            delta = tf.stack([dy, dx], axis=-1)
            # -> (1, 2r+1, 2r+1, 2)
            delta_lvl = tf.reshape(delta, (1, 2*r+1, 2*r+1, 2))

            # reshape and scale -> (bs*h*w, 1, 1, 2)
            centroid_lvl = tf.reshape(coords, (bs*h*w, 1, 1, 2)) / 2**i
            # add -> (bs*h*w, 2r+1, 2r+1, 2)
            coords_lvl = centroid_lvl + delta_lvl

            # output: (bs*h*w, 2r+1, 2r+1, dim)
            corr = bilinear_sampler(corr, coords_lvl)
            # -> (bs, h, w, (2r+1)*(2r+1)*nch)
            corr = tf.reshape(corr, (bs, h, w, -1))
            out_pyramid.append(corr)

        out = tf.concat(out_pyramid, axis=-1)
        return out

    def correlation(self, fmap1, fmap2):
        batch_size, h, w, nch = fmap1.shape
        fmap1 = tf.reshape(fmap1, (batch_size, h*w, nch))
        fmap2 = tf.reshape(fmap2, (batch_size, h*w, nch))

        # shape (batch_size, h*w, h*w)
        corr = tf.matmul(fmap1, fmap2, transpose_b=True)
        corr = tf.reshape(corr, (batch_size, h, w, 1, h, w))
        return corr / tf.sqrt(tf.cast(nch, dtype=tf.float32))        
